(* 
 * KGL scanner
 *)


{ open Parser }

let digit = ['0'-'9']
let letter = ['a'-'z' 'A'-'Z'] 
let flt = '-'? digit+ '.' digit* | '-'? '.' digit+
let identifier = letter (letter | digit | '_')*

rule token = parse

  (* Whitespace *)
  | [' ' '\t' '\n' '\r'] { token lexbuf }

  (* Comments *)
  | "/#" { comment lexbuf }
  | "##" { comment_line lexbuf }  

  (* Puncutation *)
  | '('   { LPAREN }      | ')'    { RPAREN }
  | '{'   { LBRACE }      | '}'    { RBRACE }
  | '['   { LBRACK }      | ']'    { RBRACK } 
  | ';'   { SEMI   }      | ':'    { COLON  } 
  | ','   { COMMA  }       
  | "[|"  { LLIST  }      | "|]"   { RLIST  }
  | "(|"  { LSET   }      | "|)"   { RSET   }
  | "{|"  { LGRAPH }      | "|}"   { RGRAPH } 
  | "--(" { LEDGE  }      | ")-->" { REDGE  }


  (* Arithmetic operators *)
  | '+'   { PLUS  }       | '-' { MINUS }
  | '*'   { MULTI }       | '/' { DIVIDE }
  | '%'   { MOD   }  

  (* Logical operators *)
  | "&&"  { AND }
  | "||"  { OR  }
  | "!"   { NOT }

  (* Assignment operator *)
  | "="   { ASSIGN }
  | "+="  { ADD }
  | "-="  { SUB }

  (* Relational operators *)
  | "=="  { EQUAL }       | "!=" { NEQ }
  | ">="  { GEQ }         | "<=" { LEQ }
  | '>'   { GT }          | '<'  { LT }

  (* Loop keywords *)
  | "for"       { FOR }
  | "while"     { WHILE }  
  | "continue"  { CONTINUE }
  | "break"     { BREAK }

  (* Conditional keywords *)
  | "if"        { IF }
  | "else"      { ELSE }

  (* Function keywords *)
  | "func"      { FUNC }
  | "return"    { RETURN }

  (* Type keywords *)
  | "int"       { INT   }  
  | "float"     { FLOAT }
  | "char"      { CHAR  }
  | "boolean"   { BOOL  } 
  | "string"    { STR   }
  | "graph"     { GRAPH }
  | "node"      { NODE  }
  | "list"      { LIST  } 
  | "dict"      { DICT  }


  (* Other keywords *)
  | "in"        { IN }
  | "null"      { NULL }
  | "void"      { VOID } 

  (* Literals *)
  | "true"              { BOOL_LIT(true) }
  | "false"             { BOOL_LIT(false) }
  | identifier as lit   { ID(lit) } 
  | ('0' | '-'? ['1'-'9']digit*) as lit 
                        { INT_LIT(int_of_string lit)}
  | flt as lit          { FLOAT_LIT(float_of_string lit) } 
  | '"' (([' '-'!' '#'-'[' ']'-'~'] | '\\' ['\\' '"' 'n' 't'])* as lit) '"' 
                        { STR_LIT(lit) }
  | '\'' ([^'\''] as lit) '\'' 
                        { CHAR_LIT(lit) }
  | "\'\\\\\'"          { CHAR_LIT('\\') } 
  | "\'\\t\'"           { CHAR_LIT('\t') } 
  | "\'\\n\'"           { CHAR_LIT('\n') } 

  (* Illegal characters *)
  | _ as c { raise (Failure("Illegal character: " ^ Char.escaped c)) }

  (* End of file *)
  | eof  { EOF }

and comment = parse
  | "#/" { token lexbuf }
  | _    { comment lexbuf }

and comment_line = parse
  | '\n' { token lexbuf }
  | eof  { EOF }
  | _    {comment_line lexbuf}




